
#pragma once

#include <array>
#include "State.h"

class FireDemo {
    
    public:
    
        static constexpr std::int32_t BUFFER_WIDTH = 220;
        static constexpr std::int32_t BUFFER_HEIGHT = 136;
        
        FireDemo() = delete;
        FireDemo(const FireDemo&) = delete;
        FireDemo& operator =(const FireDemo&) = delete;
        
        static inline State* state() { return _state; }
        static inline void changeState(State* state) { _next_state = state; }
    
        static inline void run(State* state) {
            _next_state = state;
            while(true) {
                if(_next_state != nullptr) {
                    if(_state != nullptr) {
                        _state->leave();
                    }
                    _state = _next_state;
                    _next_state = nullptr;
                    _state->enter();
                }
                
                _state->update();
            }
        }
        
        static constexpr const std::uint16_t* palette() { return &_palette[0]; }
        
        static constexpr std::uint32_t BGR565toIndex(std::uint32_t bgr565) {
            std::uint32_t green = (bgr565 & 0x7e0) >> 5;
            return (green > 0) ? (16 + green) : (bgr565 & 0x1f);
        }
        
        static constexpr std::uint8_t* screenBuffer() { return _screenbuf[0]; }
        
        static constexpr std::uint8_t* lineBuffer(std::uint32_t idx) { return _screenbuf[idx]; }
    
    private:
    
        static inline State* _state = nullptr;
        static inline State* _next_state = nullptr;
        
        static constexpr std::array<std::uint16_t, 80> _palette = []{
            std::array<std::uint16_t, 80> arr{};
            for(std::uint32_t idx = 0; idx < 80; ++idx) {
                arr[idx] = [idx]{
                    // Calculate RGB565 color from index
                    std::uint32_t color = (idx < 32) ? (((8 * idx) >> 3) << 11) : 0xf800;
                    if(idx >= 16) color |= ((4 * (idx - 16)) >> 2) << 5;
                    if(idx >= 48) color |= (8 * (idx - 48)) >> 3;
                    return (idx < 80) ? color : 0xffff;
                }();
            }
            return arr;
        }();
        
        static inline std::uint8_t _screenbuf[BUFFER_HEIGHT][BUFFER_WIDTH];
};
