
#include <cstdint>
#include "pomi/Display.h"
#include "pomi/Buttons.h"
#include "StateBlurFire.h"
#include "StateBufferFire.h"
#include "FireDemo.h"
#include "FixedPoint/Fixed.h"
#include "SimplexNoise.h"
#include "Random.h"

using Pomi::Display;
using Pomi::Buttons;

StateBlurFire StateBlurFire::_instance;

std::int32_t StateBlurFire::_frame;

Random StateBlurFire::_random(0);

void StateBlurFire::enter() {
    _frame = 0;
    
    Display::writeData(0); // black
    for(std::uint32_t counter = 1; counter < 220 * 176; ++counter) {
        Display::writeData();
    }
}

// 13.2 fps
void StateBlurFire::update() {
    if(Buttons::a()) {
        while(Buttons::a()) {}
        FireDemo::changeState(&StateBufferFire::instance());
        return;
    }
    
    constexpr std::int32_t WIDTH = FireDemo::BUFFER_WIDTH;
    
    constexpr const std::uint16_t* palette = FireDemo::palette();
    
    // Refresh the source of the fire at the bottom line
    Display::setX(1);
    Display::setY(176 - 1);
    Display::beginStream();
    for(std::int32_t x = 1; x < WIDTH - 1; ++x) {
        std::uint32_t envelope = ((x < WIDTH / 2) ? x : (WIDTH - x));
        
        FixedPoint::Fixed<10> xin = FixedPoint::Fixed<10>(0.05) * (x + (220 - WIDTH) / 2);
        FixedPoint::Fixed<10> yin = FixedPoint::Fixed<10>(0.02) * _frame;
        
        std::uint32_t color = 50 + (30 * SimplexNoise::noise_fixp(xin, yin)).integer();
        color = (color * ((envelope < 24) ? envelope : 24) * 2730) >> 16;
        
        Display::writeData(palette[color]);
    }
    
    std::uint8_t* buf1 = FireDemo::lineBuffer(0);
    std::uint8_t* buf2;
    
    // Initialize the first buffer with color indices from the second highest line
    Display::enterReadMode(0, 1);
    for(std::int32_t x = 0; x < WIDTH; ++x) {
        buf1[x] = FireDemo::BGR565toIndex(Display::readBGR565());
    }
    Display::leaveReadMode();
    
    for(std::int32_t y = 0; y < 176 - 2; ++y) {
        buf1 = FireDemo::lineBuffer(y & 1);    // Buffer for the line below
        buf2 = FireDemo::lineBuffer(~y & 1);   // Buffer for the line two steps below
        
        // Read the line two steps below into the second buffer
        Display::enterReadMode(0, y + 2);
        for(std::int32_t x = 0; x < WIDTH; ++x) {
            buf2[x] = FireDemo::BGR565toIndex(Display::readBGR565());
        }
        Display::leaveReadMode();
        
        // Do the fire calculations for current line and draw result on the screen
        Display::setX(1);
        Display::setY(y);
        Display::beginStream();
        for(std::int32_t x = 1; x < WIDTH - 1; ++x) {
            std::uint32_t color = (buf1[x - 1] + buf1[x + 1] + 2 * buf2[x]) / 4;
            if(color > 0) {
                Display::writeData(palette[color - _random.next(1)]);
            }
            else {
                Display::writeData(palette[0]);
            }
        }
    }
    
    // Calculate and draw the second last line separately
    Display::setX(1);
    Display::setY(176 - 2);
    Display::beginStream();
    for(std::int32_t x = 1; x < WIDTH - 1; ++x) {
        std::uint32_t color = (buf2[x - 1] + buf2[x + 1] + 2 * buf2[x]) / 4;
        if(color > 0) color -= _random.next(1);
        
        Display::writeData(palette[color]);
    }
    
    ++_frame;
}
