
#include <cstdint>
#include "pomi/Display.h"
#include "pomi/Buttons.h"
#include "StateBufferFire.h"
#include "StateDirectFire.h"
#include "FireDemo.h"
#include "FixedPoint/Fixed.h"
#include "SimplexNoise.h"
#include "Random.h"

using Pomi::Display;
using Pomi::Buttons;

StateBufferFire StateBufferFire::_instance;

std::int32_t StateBufferFire::_frame;

Random StateBufferFire::_random(0);

void StateBufferFire::enter() {
    _frame = 0;
    
    // Fill screen with diagonal black and yellow stripes
    Display::setX(0);
    Display::setY(0);
    Display::beginStream();
    for(std::uint32_t y = 0; y < 176; ++y) {
        for(std::uint32_t x = 0; x < 220; ++x) {
            Display::writeData(((x + y) & 0x10) ? 0x8400 : 0);
        }
    }
    
    std::uint8_t* screenbuf = FireDemo::screenBuffer();
    
    for(std::uint32_t idx = 0; idx < FireDemo::BUFFER_WIDTH * FireDemo::BUFFER_HEIGHT; ++idx) {
        screenbuf[idx] = 0;
    }
}

// 23.4 fps
void StateBufferFire::update() {
    if(Buttons::a()) {
        while(Buttons::a()) {}
        FireDemo::changeState(&StateDirectFire::instance());
        return;
    }
    
    constexpr std::int32_t WIDTH = FireDemo::BUFFER_WIDTH;
    constexpr std::int32_t HEIGHT = FireDemo::BUFFER_HEIGHT;
    
    constexpr const std::uint16_t* palette = FireDemo::palette();
    
    constexpr std::uint8_t* screenbuf = FireDemo::screenBuffer();
    
    // Refresh the source of the fire at the bottom line
    for(std::int32_t x = 1; x < WIDTH - 1; ++x) {
        std::uint32_t envelope = ((x < WIDTH / 2) ? x : (WIDTH - x));
        
        FixedPoint::Fixed<10> xin = FixedPoint::Fixed<10>(0.023) * (x + (220 - WIDTH) / 2);
        FixedPoint::Fixed<10> yin = FixedPoint::Fixed<10>(0.0035) * _frame;
        
        std::uint32_t color = 50 + (30 * SimplexNoise::noise_fixp(xin, yin)).integer();
        color = (color * ((envelope < 24) ? envelope : 24) * 2730) >> 16;
        
        screenbuf[WIDTH * (HEIGHT - 1) + x] = color;
    }
    
    for(std::int32_t y = 0; y < HEIGHT - 1; ++y) {
        std::uint8_t* linebuf = &screenbuf[WIDTH * (y + 1)];
        
        // Do the fire calculations for one line and draw result on the screen buffer
        for(std::uint32_t x = 1; x < WIDTH - 1; ++x) {
            std::uint32_t randn = (9 * _random.next(16)) >> 16;
            if(randn < 6) {
                std::uint8_t color = linebuf[x - 1 + (randn >> 1)];
                if(color > 0) {
                    screenbuf[WIDTH * y + x] = color - (randn & 1);
                }
                else {
                    screenbuf[WIDTH * y + x] = 0;
                }
            }
        }
    }
    
    // Copy screen buffer to the display
    Display::setX(0);
    Display::setY(176 - HEIGHT);
    Display::beginStream();
    constexpr std::uint8_t* end = &screenbuf[0] + WIDTH * HEIGHT;
    for(std::uint8_t* ptr = &screenbuf[0]; ptr < end; ++ptr) {
        Display::writeData(palette[*ptr]);
    }
    
    ++_frame;
}
