
#include <cstdint>
#include "pomi/Display.h"
#include "pomi/Buttons.h"
#include "StateDirectFire.h"
#include "StateBlurFire.h"
#include "FireDemo.h"
#include "FixedPoint/Fixed.h"
#include "SimplexNoise.h"
#include "Random.h"

using Pomi::Display;
using Pomi::Buttons;

StateDirectFire StateDirectFire::_instance;

std::int32_t StateDirectFire::_frame;

Random StateDirectFire::_random(0);

void StateDirectFire::enter() {
    _frame = 0;
    
    Display::writeData(0); // black
    for(std::uint32_t counter = 1; counter < 220 * 176; ++counter) {
        Display::writeData();
    }
}

// 12.8 fps
void StateDirectFire::update() {
    if(Buttons::a()) {
        while(Buttons::a()) {}
        FireDemo::changeState(&StateBlurFire::instance());
        return;
    }
    
    constexpr std::int32_t WIDTH = FireDemo::BUFFER_WIDTH;
    
    constexpr const std::uint16_t* palette = FireDemo::palette();
    
    // Refresh the source of the fire at the bottom line
    Display::setX(1);
    Display::setY(176 - 1);
    Display::beginStream();
    for(std::int32_t x = 1; x < WIDTH - 1; ++x) {
        std::uint32_t envelope = ((x < WIDTH / 2) ? x : (WIDTH - x));
        
        FixedPoint::Fixed<10> xin = FixedPoint::Fixed<10>(0.023) * (x + (220 - WIDTH) / 2);
        FixedPoint::Fixed<10> yin = FixedPoint::Fixed<10>(0.0035) * _frame;
        
        std::uint32_t color = 50 + (30 * SimplexNoise::noise_fixp(xin, yin)).integer();
        color = (color * ((envelope < 24) ? envelope : 24) * 2730) >> 16;
        
        Display::writeData(palette[color]);
    }
    
    std::uint8_t* linebuf = FireDemo::lineBuffer(0);
    
    for(std::int32_t y = 0; y < 176 - 1; ++y) {
        // Read the line below into the second buffer
        Display::enterReadMode(0, y + 1);
        for(std::int32_t x = 0; x < WIDTH; ++x) {
            linebuf[x] = FireDemo::BGR565toIndex(Display::readBGR565());
        }
        Display::leaveReadMode();
        
        std::uint32_t skip = 0;
        
        // Do the fire calculations for current line and draw result on the screen
        Display::setX(1);
        Display::setY(y);
        Display::beginStream();
        for(std::int32_t x = 1; x < WIDTH - 1; ++x) {
            std::uint32_t randn = (9 * _random.next(16)) >> 16;
            if(randn < 6) {
                if(skip > 0) {
                    skip = 0;
                    // Jump to current position on this line
                    Display::setX(x);
                    Display::beginStream();
                }
                
                std::uint8_t color = linebuf[x - 1 + (randn >> 1)];
                if(color > 0) {
                    Display::writeData(palette[color - (randn & 1)]);
                }
                else {
                    Display::writeData(palette[0]);
                }
            }
            else {
                ++skip;
            }
        }
    }
    
    ++_frame;
}
